import functools
import logging
import threading
from functools import wraps

from pyspark.sql import SparkSession

from sparksampling.config import SPARK_CONF
from sparksampling.error import ExhaustedError


def with_lock(func):
    # 装饰器，封装需要锁的函数
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self._acquire_lock()
        try:
            result = func(self, *args, **kwargs)
        finally:
            self._release_lock()
        return result

    return wrapper


class LogMixin(object):
    logger = logging.getLogger("sparksampling")
    log = logger


class LockMixin(object):
    # Mixin for a class need a lock
    _lock = threading.RLock()

    @classmethod
    def _acquire_lock(cls):
        """
        Acquire the class-level lock for serializing access to shared data.

        This should be released with _release_lock().
        """
        if cls._lock:
            cls._lock.acquire()

    @classmethod
    def _release_lock(cls):
        """
        Release the class-level lock acquired by calling _acquire_lock().
        """
        if cls._lock:
            cls._lock.release()


def acquire_worker(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self._allocate_worker()
        try:
            r = func(self, *args, **kwargs)
        finally:
            self._release_worker()
        return r

    return wrapper


class WorkerManagerMixin(LockMixin, LogMixin):
    # Need to set guarantee_worker before use it
    guarantee_worker = 0

    @classmethod
    @with_lock
    def _allocate_worker(cls):
        if not cls.guarantee_worker:
            raise ExhaustedError(f"No enough worker for {cls.__name__}...")
        cls.guarantee_worker = cls.guarantee_worker - 1
        cls.logger.debug(f"{cls.__name__} acquire worker:{cls.guarantee_worker} left")

    @classmethod
    @with_lock
    def _release_worker(cls):
        cls.guarantee_worker = cls.guarantee_worker + 1
        cls.logger.debug(f"{cls.__name__} release worker, have {cls.guarantee_worker} now")


def record_job_id(method):
    """
    Decorate methods
    Set job_id to Engine, Currently supports spark
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        job_id = kwargs.get("job_id", getattr(self, "job_id", None))
        if isinstance(self, SparkMixin) and job_id:
            self.log.info(f"Setting Spark job group: {job_id}")
            self.spark.sparkContext.setLogLevel("ERROR")
            self.spark.sparkContext.setJobGroup(
                job_id, f"Submitted by {self.__class__.__name__}", interruptOnCancel=True
            )
            # config for scheduler.mode: FAIR
            self.spark.sparkContext.setLocalProperty("spark.scheduler.pool", job_id)
        result = method(self, *args, **kwargs)
        return result

    return wrapper


class SparkMixin(object):
    _spark = None

    @property
    def spark(self) -> SparkSession:
        if self._spark:
            return self._spark

        if not hasattr(self, "parent"):
            return SparkSession.builder.config(conf=SPARK_CONF).getOrCreate()
        spark = getattr(self.parent, "spark")
        if not spark:
            conf = getattr(self.parent, "spark_config", SPARK_CONF)
            spark = SparkSession.builder.config(conf=conf).getOrCreate()

        self._spark = spark
        return spark



from sparksampling.engine.dummy_engine import DummyEngine
from sparksampling.evaluation.base_hook import BaseEvaluationHook


class DummyHook(BaseEvaluationHook):
    pass


DummyHook.register_pre_hook(DummyEngine)



import click
from sparksampling_client.config import reset_config
from sparksampling_client.session import Session


@click.command()
def init():
    assert Session()
    print("Pyspark sampling client initialized!")


@click.command()
def reset():
    reset_config()


@click.command()
@click.argument("json_file")
def apply(json_file):
    session = Session()
    session.apply_file(json_file)


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(reset)
cli.add_command(apply)

if __name__ == "__main__":
    cli()



import math
import random

from sparksampling.sample.base_sampling import SparkBaseSamplingJob


class ClusterSamplingImp(SparkBaseSamplingJob):
    cls_args = ["fraction", "seed", "count", "sampling_col", "group_by", "group_num"]

    def __init__(self, *args, **kwargs):
        super(ClusterSamplingImp, self).__init__(*args, **kwargs)
        self.fraction = float(kwargs.pop("fraction", 0))
        self.seed = kwargs.pop("seed", random.randint(1, 65535))
        self.count = kwargs.pop("count", 0)
        self.sampling_col = list(kwargs.pop("sampling_col", []))
        self.group_by = str(kwargs.pop("group_by"))
        self.group_num = int(kwargs.pop("group_num", 0))

    def run(self, df):
        if self.sampling_col:
            df = df[self.sampling_col]

        # https://stackoverflow.com/questions/44367019/column-name-with-dot-spark
        # Prevent .(dot) breaking select
        group = df.select(f"`{self.group_by}`").distinct()
        group_sampled = self.get_group_sampled(group)

        # equal to df.join(group_sampled, df.group_by == group_sampled.group_by, 'semi'), this will keep column in order
        cond = [getattr(df, self.group_by) == getattr(group_sampled, self.group_by)]
        df = df.join(group_sampled, cond, how="semi")
        if self.count:
            df = df.limit(self.count)

        return df

    def get_group_sampled(self, group):
        if self.fraction:
            self.log.info(f"Using fraction:{self.fraction} to sample group")
            group_df = group.sample(fraction=self.fraction, withReplacement=False, seed=self.seed)
            if self.group_num:
                self.log.info(
                    f"Fraction with group_num, limit group as group_num: {self.group_num}"
                )
                group_df = group_df.limit(self.group_num)
        else:
            # Cluster sampling expects to get the exact number of groups
            self.log.info(
                f"No fraction specified, using rdd.takeSample to get the exact number of groups"
            )
            subset = group.rdd.takeSample(False, self.group_num, seed=self.seed)
            try:
                group_df = self.spark.sparkContext.parallelize(subset).toDF()
            except ValueError as e:
                self.log.info(
                    f"Exception when rdd -> df, {e}, will use group df without being sampled"
                )
                self.log.exception(e)
                group_df = group
        return group_df



import findspark

findspark.init()

import pytest

from sparksampling.app import SparkSamplingAPP


@pytest.fixture(scope="module")
def grpc_add_to_server():
    from sparksampling.proto.sampling_service_pb2_grpc import (
        add_SparkSamplingServiceServicer_to_server,
    )
    from sparksampling.service import GRPCService

    GRPCService.get_worker_num()

    return add_SparkSamplingServiceServicer_to_server


@pytest.fixture(scope="module")
def grpc_servicer():
    from sparksampling.service import GRPCService

    SparkSamplingAPP().initialize()

    return GRPCService(SparkSamplingAPP())


@pytest.fixture(scope="module")
def grpc_stub_cls(grpc_channel):
    from sparksampling.proto.sampling_service_pb2_grpc import SparkSamplingServiceStub

    return SparkSamplingServiceStub


@pytest.fixture(scope="module", autouse=True)
def register():
    from sparksampling.engine_factory import EngineFactory

    EngineFactory.register_all_engine()


@pytest.fixture(scope="session", autouse=True)
def cleanup():
    import os
    import shutil

    dir_prefix = os.path.abspath(os.path.dirname(__file__))

    shutil.rmtree(os.path.join(dir_prefix, "./output/"), ignore_errors=True)
    shutil.rmtree(os.path.join(dir_prefix, "../data/sampled"), ignore_errors=True)

    yield



from sparksampling.engine.base_engine import BaseEngine


class DummyEngine(BaseEngine):
    guarantee_worker = 0

    @classmethod
    def register_pre_hook(cls, hook):
        cls.log.debug(
            f"Adding pre evaluation hook: {hook.__name__} to {cls.__name__}, This is a test output with no effect"
        )

    @classmethod
    def register_post_hook(cls, hook):
        cls.log.debug(
            f"Adding post evaluation hook: {hook.__name__} to {cls.__name__}, This is a test output with no effect"
        )



import os

import pytest

from sparksampling.proto.sampling_service_pb2 import (
    FILE_FORMAT_CSV,
    RANDOM_SAMPLING_METHOD,
    FileFormatConf,
    RelationSamplingRequest,
    SamplingConf,
)

dir_prefix = os.path.abspath(os.path.dirname(__file__))
input_path = os.path.join(dir_prefix, "../data/10w_x_10.csv")
input_a = os.path.join(dir_prefix, "../data/relation-a.csv")
input_b = os.path.join(dir_prefix, "../data/relation-b.csv")
input_c = os.path.join(dir_prefix, "../data/relation-c.csv")
SEED = 4


def test_only_sampling(grpc_stub):
    sampling_stage = RelationSamplingRequest.SamplingStage(
        input_path=input_a,
        output_path=os.path.join(dir_prefix, "./output/only-sampling/a-sampled"),
        name="sampling-stage",
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=SamplingConf(
            fraction="0.5",
            with_replacement=True,
            seed=SEED,
        ),
        format_conf=FileFormatConf(with_header=True, sep=","),
    )
    request = RelationSamplingRequest(
        sampling_stages=[sampling_stage],
        job_id="only-sampling-job",
    )

    response = grpc_stub.RelationSamplingJob(request)
    assert response.code == 0


def test_only_relation(grpc_stub):
    relation_b = RelationSamplingRequest.RelationStage(
        input_path=input_b,
        output_path=os.path.join(dir_prefix, "./output/only-relation/b-sampled"),
        name="b",
        relations=[
            RelationSamplingRequest.Relation(
                relation_path=input_a, relation_col="user_id", input_col="user_id"
            )
        ],
    )

    relation_c = RelationSamplingRequest.RelationStage(
        input_path=input_c,
        output_path=os.path.join(dir_prefix, "./output/only-relation/c-sampled"),
        name="c",
        relations=[
            RelationSamplingRequest.Relation(
                relation_name="b", relation_col="session_id", input_col="session_id"
            )
        ],
    )

    request = RelationSamplingRequest(
        sampling_stages=[],
        relation_stages=[relation_b, relation_c],
        job_id="only-relation-job",
        default_sampling_method=RANDOM_SAMPLING_METHOD,
        default_file_format=FILE_FORMAT_CSV,
        default_sampling_conf=SamplingConf(
            fraction="0.5",
            with_replacement=True,
            seed=SEED,
        ),
        default_format_conf=FileFormatConf(with_header=True, sep=","),
    )

    response = grpc_stub.RelationSamplingJob(request)
    assert response.code == 0


def test_part_relation(grpc_stub):
    sampling_a = RelationSamplingRequest.SamplingStage(
        input_path=input_a,
        output_path=os.path.join(dir_prefix, "./output/part-relation/a-sampled"),
        name="a",
    )
    relation_b = RelationSamplingRequest.RelationStage(
        input_path=input_b,
        output_path=os.path.join(dir_prefix, "./output/part-relation/b-sampled"),
        name="b",
        relations=[
            RelationSamplingRequest.Relation(
                relation_path=input_a,
                relation_col="user_id",
            )
        ],
    )

    relation_c = RelationSamplingRequest.RelationStage(
        input_path=input_c,
        output_path=os.path.join(dir_prefix, "./output/part-relation/c-sampled"),
        name="c",
        relations=[
            RelationSamplingRequest.Relation(
                relation_path=input_b,
                relation_col="session_id",
            )
        ],
    )

    request = RelationSamplingRequest(
        sampling_stages=[sampling_a],
        relation_stages=[relation_b, relation_c],
        job_id="part-relation-job",
        default_sampling_method=RANDOM_SAMPLING_METHOD,
        default_file_format=FILE_FORMAT_CSV,
        default_sampling_conf=SamplingConf(
            fraction="0.5",
            with_replacement=True,
            seed=SEED,
        ),
        default_format_conf=FileFormatConf(with_header=True, sep=","),
    )

    response = grpc_stub.RelationSamplingJob(request)
    assert response.code == 0


def test_chain_relation(grpc_stub):
    sampling_a = RelationSamplingRequest.SamplingStage(
        input_path=input_a,
        output_path=os.path.join(dir_prefix, "./output/chain-relation/a-sampled"),
        name="a",
    )
    relation_b = RelationSamplingRequest.RelationStage(
        input_path=input_b,
        output_path=os.path.join(dir_prefix, "./output/chain-relation/b-sampled"),
        name="b",
        relations=[
            RelationSamplingRequest.Relation(
                relation_name="a", relation_col="user_id", input_col="user_id"
            )
        ],
    )

    relation_c = RelationSamplingRequest.RelationStage(
        input_path=input_c,
        output_path=os.path.join(dir_prefix, "./output/chain-relation/c-sampled"),
        name="c",
        relations=[
            RelationSamplingRequest.Relation(
                relation_name="b", relation_col="session_id", input_col="session_id"
            )
        ],
    )

    request = RelationSamplingRequest(
        sampling_stages=[sampling_a],
        relation_stages=[relation_b, relation_c],
        job_id="chain-relation-job",
        default_sampling_method=RANDOM_SAMPLING_METHOD,
        default_file_format=FILE_FORMAT_CSV,
        default_sampling_conf=SamplingConf(
            fraction="0.5",
            with_replacement=True,
            seed=SEED,
        ),
        default_format_conf=FileFormatConf(with_header=True, sep=","),
    )

    response = grpc_stub.RelationSamplingJob(request)
    assert response.code == 0


def test_dry_run(grpc_stub):
    out_path = os.path.join(dir_prefix, "./output/dry/a-sampled")
    sampling_a = RelationSamplingRequest.SamplingStage(
        input_path=input_a,
        output_path=out_path,
        name="a",
    )

    request = RelationSamplingRequest(
        sampling_stages=[sampling_a],
        relation_stages=[],
        job_id="dry-run-job",
        default_sampling_method=RANDOM_SAMPLING_METHOD,
        default_file_format=FILE_FORMAT_CSV,
        default_sampling_conf=SamplingConf(
            fraction="0.5",
            with_replacement=True,
            seed=SEED,
        ),
        default_format_conf=FileFormatConf(with_header=True, sep=","),
        dry_run=True,
    )

    response = grpc_stub.RelationSamplingJob(request)
    assert response.code == 0
    assert not os.path.exists(out_path)


if __name__ == "__main__":
    pytest.main()



from os.path import abspath, dirname, exists, join, split
from pprint import pformat
from typing import Dict
from uuid import uuid4

from sparksampling.engine.base_engine import SparkBaseEngine
from sparksampling.engine_factory import EngineFactory
from sparksampling.error import BadParamError, CustomErrorWithCode, ProcessError
from sparksampling.file_format.file_factory import FileFormatFactory
from sparksampling.mixin import acquire_worker, record_job_id
from sparksampling.proto.sampling_service_pb2 import SamplingRequest
from sparksampling.sample import SamplingFactory
from sparksampling.utilities import check_spark_session


class SMSEngine(SparkBaseEngine):
    # Single Mapping Sampling

    def __init__(
        self,
        parent,
        input_path: str,
        output_path: str,
        sampling_method: int,
        file_format: int,
        job_id: str,
        sampling_conf: Dict or None = None,
        format_conf: Dict or None = None,
    ):
        super(SMSEngine, self).__init__()
        self.parent = parent
        self.input_path = input_path
        self.output_path = output_path
        self.sampling_method = sampling_method
        self.file_format = file_format
        self.job_id = job_id
        self.sampling_conf = sampling_conf
        self.format_conf = format_conf

    @acquire_worker
    @check_spark_session
    def submit(self, *args, **kwargs):
        sampling_imp, file_imp = self.get_spark_imp()
        return self.submit_spark_job(sampling_imp, file_imp)

    def get_spark_imp(self):
        try:
            sampling_imp = SamplingFactory.get_sampling_imp(
                self.sampling_method, self.sampling_conf
            )
            file_imp = FileFormatFactory.get_file_imp(
                self.spark, self.file_format, self.format_conf
            )
        except CustomErrorWithCode as e:
            raise e
        except (KeyError, ValueError) as e:
            self.log.info(f"Task initialization failed {e}")
            raise BadParamError(f"Check task configuration parameters {str(e)}")
        except Exception as e:
            self.log.exception(e)
            raise ProcessError(str(e))
        return sampling_imp, file_imp

    @record_job_id
    def submit_spark_job(self, sampling_imp, file_imp):
        df = file_imp.read(self.input_path)
        df, pre_metas = self.pre_hook(df)
        df = sampling_imp.run(df)
        df, post_metas = self.post_hook(df)
        output_path = file_imp.write(df, self.output_path)
        self.log.info(f"The task is completed and the output file or directory is: {output_path}")
        return output_path, pre_metas + post_metas

    @classmethod
    def config(cls, kwargs):
        def _set_default_output_path(conf):
            if exists(conf.get("input_path")):
                # when input_path = /{path_to_data}/{input_file_name}
                # set default_output_path = /{path_to_data}/sampled/{input_file_name}-{uuid}
                input_path = conf.get("input_path")
                default_output_path = abspath(
                    join(
                        dirname(abspath(input_path)),
                        "./sampled/",
                        split(input_path)[-1] + "-" + str(uuid4()),
                    )
                )
                conf.setdefault("output_path", default_output_path)
            else:
                input_path = conf.get("input_path")
                default_output_path = input_path.rstrip("/") + "-" + str(uuid4())
                conf.setdefault("output_path", default_output_path)
            return conf

        try:
            kwargs = _set_default_output_path(kwargs)
            required_conf = {
                "input_path": kwargs.pop("input_path"),
                "output_path": kwargs.pop("output_path"),
                "sampling_method": kwargs.pop("sampling_method"),
                "file_format": kwargs.pop("file_format"),
                "job_id": kwargs.pop("job_id"),
            }
        except KeyError as e:
            cls.log.info(f"Missing required parameters {str(e)}")
            raise BadParamError(f"Missing required parameters {str(e)}")

        # Detailed configuration depends on the specific implementation
        conf = {
            "sampling_conf": kwargs.get("sampling_conf", {}),
            "format_conf": kwargs.get("format_conf", {}),
        }
        conf.update(required_conf)
        cls.log.info(f"Initializing job conf... \n {pformat(conf)}")
        return conf

    @classmethod
    def is_matching(cls, request_type):
        return request_type == SamplingRequest


EngineFactory.register(SMSEngine)



import json
import os
import random
from typing import Dict

from pyspark.sql.functions import col, rand, row_number
from pyspark.sql.window import Window

from sparksampling.sample.base_sampling import SparkBaseSamplingJob


class StratifiedSamplingImp(SparkBaseSamplingJob):
    cls_args = [
        "fraction",
        "with_replacement",
        "stratified_key",
        "seed",
        "count",
        "sampling_col",
        "ensure_col",
    ]

    def __init__(self, *args, **kwargs):
        super(StratifiedSamplingImp, self).__init__(*args, **kwargs)
        self.fraction = self._init_fraction(kwargs.pop("fraction"))
        self.stratified_key = kwargs.pop("stratified_key")
        self.with_replacement = kwargs.pop("with_replacement", False)
        self.seed = kwargs.pop("seed", random.randint(1, 65535))
        self.count = kwargs.pop("count", 0)
        self.sampling_col = list(kwargs.pop("sampling_col", []))
        self.ensure_col = os.environ.get("FORCE_STRATIFILED_ENSURE_COL") in [
            "True",
            "true",
        ] or kwargs.pop("ensure_col", False)

    @staticmethod
    def _init_fraction(fraction_str):
        fraction = json.loads(fraction_str)
        if isinstance(fraction, int):
            fraction = float(fraction)
        if not isinstance(fraction, (Dict, float)):
            raise ValueError("fraction must be float or dict")
        return fraction

    def run(self, df):
        fraction = self._convert_fraction_to_dict(df, self.stratified_key, self.fraction)
        if self.sampling_col:
            df = df[self.sampling_col]
        sampled_df = df.sampleBy(col=self.stratified_key, fractions=fraction, seed=self.seed)
        if self.ensure_col:
            self.log.info("ensure every layer in result, ignore count")
            sampled_df = self.sample_full_layer_df(sampled_df, df, self.stratified_key)
        if self.count:
            sampled_df = sampled_df.limit(self.count)
        return sampled_df

    @staticmethod
    def _convert_fraction_to_dict(df, stratified_key, fraction) -> Dict:
        if isinstance(fraction, Dict):
            return fraction

        # https://stackoverflow.com/questions/44367019/column-name-with-dot-spark
        # Prevent .(dot) breaking select
        StratifiedSamplingImp.log.info(f"Convert fraction {fraction} to dict...")
        y = df.select(f"`{stratified_key}`")
        convert_fraction = {label[0]: fraction for label in y.distinct().toLocalIterator()}
        return convert_fraction

    @staticmethod
    def sample_full_layer_df(sampled_df, origin_df, stratified_key):
        sampled_layer_key = sampled_df.select(f"`{stratified_key}`").distinct()
        missing_layer_key = (
            origin_df.select(f"`{stratified_key}`").distinct().exceptAll(sampled_layer_key)
        )

        missing_df = origin_df.join(missing_layer_key, stratified_key, "semi")
        stratified_window = Window.partitionBy(f"`{stratified_key}`").orderBy(rand())
        missing_layer = (
            missing_df.withColumn("_", row_number().over(stratified_window))
            .filter(col("_") == 1)
            .drop("_")
        )
        return sampled_df.unionByName(missing_layer, allowMissingColumns=True)



# import all .py from in this dir
import glob
import importlib
import os
from os.path import basename, dirname, isfile, join

modules = glob.glob(join(dirname(__file__), "*.py"))
sub_packages = (basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py"))
packages = (str(__package__) + "." + i for i in sub_packages)
[importlib.import_module(p) for p in packages]

evaluation_module_name = "evaluation_extension"
evaluation_extension_path = os.path.join(os.path.dirname(__file__), f"../{evaluation_module_name}")
evaluation_sub_folders = (f.name for f in os.scandir(evaluation_extension_path) if f.is_dir())
sub_packages = (f for f in evaluation_sub_folders if not (f.endswith("__") or f.startswith("__")))
evaluation_package = ".".join((str(__package__).split("."))[:-1])
packages = [evaluation_package + "." + evaluation_module_name + "." + i for i in sub_packages]
[importlib.import_module(p) for p in packages]



import os
import weakref

from sparksampling.error import ProcessPostHookError, ProcessPreHookError
from sparksampling.mixin import SparkMixin, WorkerManagerMixin


class BaseEngine(WorkerManagerMixin):
    guarantee_worker = int(os.getenv("ENGINE_DEFAULT_WORKER_NUM", 10))
    evaluation_pre_hook = dict()
    evaluation_post_hook = dict()
    _cache_hook_instance = weakref.WeakValueDictionary()

    def submit(self, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def stop(cls, parent, job_id=None):
        raise NotImplementedError

    @classmethod
    def config(cls, kwargs):
        raise NotImplementedError

    @classmethod
    def is_matching(cls, request_type):
        return False

    @classmethod
    def register_pre_hook(cls, hook):
        evaluation_pre_hook = cls.evaluation_pre_hook.setdefault(cls, set())
        if hook in evaluation_pre_hook:
            return
        cls.log.info(f"Adding pre evaluation hook: {hook.__name__} to {cls.__name__}")
        evaluation_pre_hook.add(hook)

    @classmethod
    def register_post_hook(cls, hook):
        evaluation_post_hook = cls.evaluation_post_hook.setdefault(cls, set())

        if hook in evaluation_post_hook:
            return
        cls.log.info(f"Adding post evaluation hook: {hook.__name__} to {cls.__name__}")
        evaluation_post_hook.add(hook)

    @classmethod
    def _process_hook(cls, df, hooks, exception, period):
        metas = []
        for hook in hooks:
            try:
                df, meta = cls.get_hook_instance(hook).process(df)
                metas.append(meta.generate_proto_msg(period))
            except NotImplementedError:
                raise exception(
                    f"{period} hook raised, Not implemented hook:{hook} found in {hooks}"
                )
            except Exception as e:
                cls.log.info(
                    f"Exception when processing df: {df} with hook: {hook} in period {period}"
                )
                cls.logger.exception(e)
                raise exception(str(e))

        return df, metas

    @classmethod
    def pre_hook(cls, df):
        return cls._process_hook(
            df, cls.evaluation_pre_hook.get(cls, set()), ProcessPreHookError, "pre"
        )

    @classmethod
    def post_hook(cls, df):
        return cls._process_hook(
            df, cls.evaluation_post_hook.get(cls, set()), ProcessPostHookError, "post"
        )

    @classmethod
    def get_hook_instance(cls, hook):
        return cls._cache_hook_instance.setdefault(hook, hook())


class SparkBaseEngine(BaseEngine, SparkMixin):
    @classmethod
    def stop(cls, parent, job_id=None):
        # SparkStopEngine will cancel spark job
        return



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

requirements = ["sparksampling"]

setup(
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="sparksampling-evaluation-extension-example",
    install_requires=requirements,
    include_package_data=True,
    keywords="sparksampling-evaluation-extension-example",
    name="sparksampling-evaluation-extension-example",
    packages=find_packages(include=["sparksampling.evaluation_extension.example_extension"]),
    version="0.1.0",
    zip_safe=False,
)



import os

import findspark

findspark.init()

import logging
from concurrent import futures

import grpc
from pyspark import SparkConf
from pyspark.sql import SparkSession
from traitlets import Instance, Integer, Unicode, default
from traitlets.config import Application

from sparksampling import __version__
from sparksampling.config import SPARK_CONF
from sparksampling.service import GRPCService


class SparkSamplingAPP(Application):
    name = "sparksampling"
    description = """An application for starting a spark sampling server"""
    # the grpc server handle
    server = None

    ip = Unicode(
        os.getenv("SERVICE_HOST", "0.0.0.0"),
        help="Host IP address for listening (default 0.0.0.0).",
    ).tag(config=True)

    port = Integer(int(os.getenv("SERVICE_PORT", 8530)), help="Port (default 8530).").tag(
        config=True
    )

    spark_config = Instance(SparkConf)
    spark = Instance(SparkSession)

    config_file_path = Unicode(
        os.path.expanduser("~/.ss/config.py"), help="User defined app config path"
    ).tag(config=True)

    @default("spark_config")
    def _spark_config_default(self):
        return SPARK_CONF

    @default("spark")
    def _spark_default(self):
        return SparkSession.builder.config(conf=self.spark_config).getOrCreate()

    @default("log_level")
    def _log_level_default(self):
        return logging.DEBUG

    @default("log_datefmt")
    def _log_datefmt_default(self):
        """Exclude date from default date format"""
        return "%Y-%m-%d %H:%M:%S"

    @default("log_format")
    def _log_format_default(self):
        """override default log format to include time"""
        return "[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] %(message)s"

    def load_config(self):
        dir_prefix = os.path.abspath(os.path.dirname(__file__))
        self.load_config_file(os.path.join(dir_prefix, "default_app_config.py"))
        self.log.info(f"Loading user defined config from {self.config_file_path}")
        self.load_config_file(self.config_file_path)

    def initialize(self, *args, **kwargs):
        self.load_config()
        super().initialize(*args, **kwargs)
        self.init_logger()
        self.log.info(f"Current pyspark-sampling version: {__version__}")
        self.init_spark()

    def init_logger(self):
        logger = logging.getLogger("sparksampling")
        logger.propagate = True
        logger.parent = self.log
        logger.setLevel(self.log.level)

    def init_spark(self):
        self.log.info(f"Started SparkSession, Spark version: {self.spark.version}")

    def start(self, argv=None):
        self.initialize(argv)
        self._add_server()
        self.server.add_insecure_port("%s:%d" % (self.ip, self.port))
        self.server.start()
        self.log.info("Spark Sampling Server Listening On %s:%s..." % (self.ip, self.port))
        self.server.wait_for_termination()

    def _add_server(self):
        GRPCService.register_engine(self)
        engine_worker = GRPCService.get_worker_num()
        # need one reserve worker to reject request asap
        max_workers = engine_worker + 1
        self.log.info(
            f"Service allocate {engine_worker} workers for engine and 1 reserve for response"
        )
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        GRPCService.add_to_server(self, self.server)

    @classmethod
    def launch(cls, argv=None):
        self = cls.instance()
        self.start(argv)


main = SparkSamplingAPP.launch

if __name__ == "__main__":
    main()



import json
import os
import pathlib
import subprocess

import pytest
from sparksampling_client.proto.sampling_service_pb2 import (
    FILE_FORMAT_CSV,
    RANDOM_SAMPLING_METHOD,
)


@pytest.fixture(scope="module")
def request_dict():
    sampling_col = ["# id", "X_0", "X_1", "X_2", "y"]
    dir_prefix = os.path.abspath(os.path.dirname(__file__))
    input_path = os.path.join(dir_prefix, "../../data/10w_x_10.csv")
    output_path = os.path.join(dir_prefix, "./output/test_random_sampling_fraction")

    return dict(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=dict(fraction="0.2", with_replacement=False, sampling_col=sampling_col),
        format_conf=dict(with_header=True, sep=","),
        input_path=input_path,
        output_path=output_path,
        job_id="test_random_sampling",
    )


@pytest.fixture
def json_request_file(tmpdir, request_dict):
    json_file = pathlib.Path(tmpdir / "reqeust.json")
    with json_file.open("w") as f:
        json.dump(request_dict, f)
    return json_file


def test_apply(json_request_file):
    subprocess.run(["ssc", "apply", json_request_file], check=True)


if __name__ == "__main__":
    pytest.main()



import json
import os
import pathlib

CONFIG_PATH = os.path.expanduser("~/.ssc/config.json")
CONFIG_FILE = pathlib.Path(CONFIG_PATH)
CONFIG_FILE.parent.mkdir(exist_ok=True)
DEFAULT_CONFIG = {
    "host": "localhost",
    "port": "8530",
}


def load_config():
    if not CONFIG_FILE.exists():
        print(f"No config file found, generating default config in {CONFIG_FILE}")
        with CONFIG_FILE.open("w") as f:
            json.dump(DEFAULT_CONFIG, f)
    print(f"Loading config from {CONFIG_FILE}")
    with CONFIG_FILE.open() as f:
        return json.load(f)


def reset_config():
    print(f"Config {CONFIG_FILE} reset to default")
    with CONFIG_FILE.open("w") as f:
        json.dump(DEFAULT_CONFIG, f)



from sparksampling.engine.base_engine import SparkBaseEngine
from sparksampling.engine_factory import EngineFactory


class SparkStopEngine(SparkBaseEngine):
    guarantee_worker = 1

    @classmethod
    def stop(cls, parent, job_id=None):
        cls.log.info(f"Send Stop Job to Spark: {job_id}")
        parent.spark.sparkContext.cancelJobGroup(job_id)


EngineFactory.register(SparkStopEngine)



import random

from sparksampling.sample.base_sampling import SparkBaseSamplingJob


class RandomSamplingImp(SparkBaseSamplingJob):
    cls_args = ["fraction", "with_replacement", "seed", "count", "sampling_col"]

    def __init__(self, *args, **kwargs):
        super(RandomSamplingImp, self).__init__(*args, **kwargs)
        self.fraction = float(kwargs.pop("fraction", 0))
        self.with_replacement = kwargs.pop("with_replacement", False)
        self.seed = kwargs.pop("seed", random.randint(1, 65535))
        self.count = kwargs.pop("count", 0)
        self.sampling_col = list(kwargs.pop("sampling_col", []))

    def run(self, df):
        if self.sampling_col:
            df = df[self.sampling_col]

        # Obtain 5% more samples to meet count requirements
        fraction = min(self.fraction or 1.05 * (self.count / df.count()), 1.0)
        self.log.info(f"Start sampling as fraction: {fraction}")
        df = df.sample(fraction=fraction, seed=self.seed, withReplacement=self.with_replacement)
        if self.count:
            df = df.limit(self.count)
        return df
